/*  mmriseg.h

    Yongyue Zhang, FMRIB Image Analysis Group
	
	Y. Zhang, M. Brady, S. Smith, "Segmentation of Brain MR Images Through 
	a Hidden Markov Random Field Model and the Expectation Maximization 
	Algorithm", IEEE Trans. Medical Imaging, Vol. 20 No. 1, pp. 45-57. 2001.

    Copyright (C) 1999-2001 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

#ifndef __MULTIMRISEGMENTATION__H_
#define __MULTIMRISEGMENTATION__H_

#include <vector>
#include "imagecore.h"

// multi-spectral version of ZMRISegmentation
class ZMultiMRISegmentation
{
	enum VERBOSE { QUIET=0, MAIN=1, SLICE=2, ITER=3, PARA=4, PROGRESS=5 };

	int		m_nChannel;

	int		m_nWidth;			// image width
	int		m_nHeight;			// image height
	int		m_nDepth;			// image depth (number of slices)
	int		m_nSliceSize;		// width x height
	int		m_nSize;			// m_nSliceSize x m_nDepth
	
	int		m_nBegin;			// the first slice to be used for parameter estimation
	int		m_nEnd;				// the last slice to be used for parameter estimation

	float	pixdim[3];			// pixel dimmension in mm.

	bool	m_b3DLowpass;		// whether to do 3D lowpass filtering

	VERBOSE	m_nVerbose;			// verbose mode 

	BYTE	neighbour[27];		// for MRF neighbours of a pixel (second-order)
	float	B;					// MRF beta parameter
	float	diaB;				// MRF beta parameter used for diagonal neighbours
	float	ddiaB;				// MRF beta parameter used for neighbours in different slices
	float	TB;					// maximum prior of a pixel
	int		offset3d;			// offset of the first pixel in a 3x3x3 cube with respect to the center pixel
	int		offset2d;			// offset of the first pixel in a 3x3 cube with respect to the center pixel

	int		m_left, m_right, m_top, m_bottom, m_front, m_back;	// ROI of the images

	const std::vector<ZImageBase*>* m_images;

	std::vector<BYTE>	m_mask;

	ZGrayFloatImage		*m_FloatImage;

	void			EMloop(void);
	int ICM(float BB);
	void Lowpass(ZMatrix<float>& meanres, std::vector< ZMatrix<float> >& meaninvcov);
	void Initialize();
	// determine the range of slices for parameter estimation
	bool	DetermineValidSlices();

//////////////////////////////////////////////////////
// PVE members
//////////////////////////////////////////////////////
	void	PVE();			// PVE classification
	void	InitializePVE();
	float	ComputeConditional(const ZVector<float>& mri, const ZVector<float>& pve);
	float	ComputePosterior(const ZVector<float>* p_mri, const ZVector<float>* p_pve, 
				std::vector<BYTE>::const_iterator p_mask, const ZVector<float>& rpv);
public:

	ZMatrix<float>			m_Mri, m_Bias;		//m_nSize x m_nChannel
	ZGrayByteImage			m_Segment;
	
	ZMatrix<float> m_prob;			// likelihood probability map
	ZMatrix<float> m_post;			// posterior probability map
	ZVector<float> m_prior;
	ZMatrix<BYTE> m_apriori;		// apriori probability map

	int		m_nbClasses;			// number of classes
	int		m_nbIter;				// number of iterations
	int		m_nbLowpass;			// number of lowpass filtering for bias field

	UINT	m_nSampleRatio;			// the sample ratio of when doing parameter updating

	bool	m_b3D;					// 3D segmentation or 2D segmentation
	bool	m_b2DImage;				// the original image is a 2D image
	bool	m_bBias;				// whether to correct for the bias field

	ZMatrix<float>  m_mean;			// m_nbClasses x m_nChannel
	// m_nChannel x m_nChannel
	std::vector< ZMatrix<float> >	m_covariance, m_icovariance;		
	ZVector<float>			m_stddev;	// 1/||m_covariance||^2
	ZVector<UINT>			m_number;

	ZMatrix<float>	m_inimean;
	std::vector< ZMatrix<float> >	m_inicovariance;

	ZMultiMRISegmentation();
	~ZMultiMRISegmentation() { delete m_FloatImage; }
	
	void Create(const std::vector<ZImageBase*>& images, int nclass, bool b2d,
				int nbiter,	int nblowpass, float neighbor, UINT sampleratio, bool bBias=true);

	bool Segment(bool pve, const ZGrayByteImage* gm=0, const ZGrayByteImage* wm=0, const ZGrayByteImage* csf=0);

	ZGrayFloatImage* GetRestored(int channel);
	ZGrayFloatImage* GetBiasField(int channel);
	ZGrayFloatImage* GetProbability(int wclass);

	void SetVerboseMode(int verbose=1)  
	{ 
		if(verbose<0 || verbose>5) verbose = 1;
		m_nVerbose = VERBOSE(verbose); 
	}

//////////////////////////////////////////////////////
// PVE members
//////////////////////////////////////////////////////
	ZVector<float>	tv;				// temporary vector (m_nbClasses)
	ZMatrix<float>	m_PVE;			// m_nSize x m_nbClasses
	ZVector<float>	m_energy;		// energy
	ZGrayFloatImage m_PVEimage;		// PVE image
	bool			m_bPVEPara;		// whether updating parameters during PVE
	float			PVEB;			// beta for PVE MRF

	void SetPVEPara(bool bPara, float beta) { m_bPVEPara = bPara; PVEB = beta; }
	ZGrayFloatImage* GetPVE(int wclass);
};

#endif
