/*  miscmaths.cc

    Mark Jenkinson, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

// Miscellaneous maths functions

#include "miscmaths.h"
#include "stdlib.h"

namespace MISCMATHS {

  // The following lines are ignored by the current SGI compiler
  //  (version egcs-2.91.57)
  // A temporary fix of including the std:: in front of all abs() etc
  //  has been done for now
  using std::abs;
  using std::sqrt;
  using std::exp;
  using std::log;
  using std::pow;

  // General string/IO functions

  bool isnum(const string& str)
  {
    // assumes that initial whitespace has been removed
    if (isdigit(str[0])) return true;
    if ( (str[0]=='-') || (str[0]=='+') || (str[0]=='.') )  return true;
    return false;
  }
    
  string skip_alpha(ifstream& fs) 
  {
    string cline;
    while (!fs.eof()) {
      getline(fs,cline);
      cline += " "; // force extra entry in parsing
      istrstream ss(cline.c_str());
      string cc="";
      ss >> cc;
      if (isnum(cc)) {
	fs.seekg(-((int)cline.size()),ios::cur);
	return cline;
      }
    }
    return "";
  }

  Matrix read_ascii_matrix(int nrows, int ncols, const string& filename)
  {
    return read_ascii_matrix(filename,nrows,ncols);
  }


  Matrix read_ascii_matrix(const string& filename, int nrows, int ncols)
  {
    Matrix mat(nrows,ncols);
    mat = 0.0;
  
    if ( filename.size()<1 ) return mat;
    ifstream fs(filename.c_str());
    if (!fs) { 
      cerr << "Could not open matrix file " << filename << endl;
      return mat;
    }
    mat = read_ascii_matrix(fs,nrows,ncols);
    fs.close();
    return mat;
  }

  Matrix read_ascii_matrix(int nrows, int ncols, ifstream& fs)
  {
    return read_ascii_matrix(fs, nrows, ncols);
  }

  Matrix read_ascii_matrix(ifstream& fs, int nrows, int ncols)
  {
    Matrix mat(nrows,ncols);
    mat = 0.0;
    string ss="";
  
    ss = skip_alpha(fs);
    for (int r=1; r<=nrows; r++) {
      for (int c=1; c<=ncols; c++) {
	if (!fs.eof()) {
	  fs >> ss;
	  while ( !isnum(ss) && !fs.eof() ) {
	    fs >> ss;
	  }
	  if (!fs.eof()) mat(r,c) = atof(ss.c_str());
	}
      }
    }
    return mat;
  }

  
  Matrix read_ascii_matrix(const string& filename)
  {
    Matrix mat;
    if ( filename.size()<1 ) return mat;
    ifstream fs(filename.c_str());
    if (!fs) { 
      cerr << "Could not open matrix file " << filename << endl;
      return mat;
    }
    mat = read_ascii_matrix(fs);
    fs.close();
    return mat;
  }


  Matrix read_ascii_matrix(ifstream& fs)
  {
    Matrix mat;

    int rcount=0, cmax=0;
    double val;
    string cline;
    // skip initial non-numeric lines
    //  and count the number of columns in the first numeric line
    cline = skip_alpha(fs);
    cline += " ";
    {
      istrstream ss(cline.c_str());
      string cc="";
      while (!ss.eof()) {
	cmax++;
	ss >> cc;
      }
    }
    cmax--;
    RowVector newrow(cmax);

    while (!fs.eof()) {
      getline(fs,cline);
      cline += " "; // force extra entry in parsing
      istrstream ss(cline.c_str());
      string cc="";
      ss >> cc;
      if (!isnum(cc)) break;  // stop processing when non-numeric line found
      rcount++;  // add new row to matrix
      newrow = 0.0;
      int ccount = 1;
      do {
	val = atof(cc.c_str());
	if (ccount<=cmax)  newrow(ccount++) = val;
	ss >> cc;
      } while (!ss.eof());
  
      if (rcount>1) mat = mat & newrow;
      else mat = newrow;
    }
  
    return mat;
  }

  int write_ascii_matrix(const string& filename, const Matrix& mat, 
			 int precision)
  {
    return write_ascii_matrix(mat, filename, precision);
  }

  int write_ascii_matrix(const Matrix& mat, const string& filename, 
			 int precision)
  {
    Tracer tr("write_ascii_matrix");
    if ( (filename.size()<1) ) return -1;
    ofstream fs(filename.c_str());
    if (!fs) { 
      cerr << "Could not open file " << filename << " for writing" << endl;
      return -1;
    }
    int retval = write_ascii_matrix(mat,fs,precision);
    fs.close();
    return retval;
  }

  int write_ascii_matrix(ofstream& fs, const Matrix& mat, 
			 int precision)
  {
    return write_ascii_matrix(mat, fs, precision);
  }
  
  int write_ascii_matrix(const Matrix& mat, ofstream& fs, int precision)
  {
    for (int i=1; i<=mat.Nrows(); i++) {
      for (int j=1; j<=mat.Ncols(); j++) {
	if (precision>0)  { 
	  fs.setf(ios::scientific | ios::showpos); 
	  fs.precision(precision+7);  
	  fs.precision(precision); } 
	fs << mat(i,j) << "  ";
      }
      fs << endl;
    }
    return 0;
  }


  // General mathematical functions

  int round(int x) { return x; }

  int round(float x) 
    { 
      if (x>0.0) return ((int) (x+0.5));
      else       return ((int) (x-0.5));
    }

   int round(double x) 
   { 
     if (x>0.0) return ((int) (x+0.5));
     else       return ((int) (x-0.5));
   }
   
   complex power(const complex& a, const complex& b)
   { 
     complex res(0,0);
     if (abs(a)==0.0) { return res; }
     else { return pow(a,b); }
   }

   int periodicclamp(int x, int x1, int x2)
   {
     if (x2<x1) return periodicclamp(x,x2,x1);
     int d = x2-x1+1;
     int xp = x-x1;
     if (xp>=0) {
       return (xp % d) + x1;
     } else {
       xp = xp + d + std::abs(xp/d)*d;
       assert(xp>0);
       return periodicclamp(xp + d + std::abs(xp/d)*d,x1,x2);
     }
   }

  ColumnVector cross(const ColumnVector& a, const ColumnVector& b)
    {
      Tracer tr("cross");
      ColumnVector ans(3);
      ans(1) = a(2)*b(3) - a(3)*b(2);
      ans(2) = a(3)*b(1) - a(1)*b(3);
      ans(3) = a(1)*b(2) - a(2)*b(1);
      return ans;
    }


  ColumnVector cross(const Real *a, const Real *b)
    {
      Tracer tr("cross");
      ColumnVector a1(3), b1(3);
      a1 << a;
      b1 << b;
      return cross(a1,b1);
    }


  double norm2(const ColumnVector& x)
    {
      return std::sqrt(x.SumSquare());
    }


  int Identity(Matrix& m)
    {
      Tracer tr("Identity");
      m=0.0;
      for (int j=1; j<=m.Nrows(); j++)
	m(j,j)=1.0;
      return 0;
    }

  Matrix Identity(int num)
    {
      Tracer tr("Identity");
      Matrix eye(num,num);
      Identity(eye);
      return eye;
    }


  int diag(Matrix& m, const float diagvals[])
    {
      Tracer tr("diag");
      m=0.0;
      for (int j=1; j<=m.Nrows(); j++)
	m(j,j)=diagvals[j-1];
      return 0;
    }


  int diag(Matrix& m, const ColumnVector& diagvals)
    {
      Tracer tr("diag");
      m=0.0;
      for (int j=1; j<=m.Nrows(); j++)
	m(j,j)=diagvals(j);
      return 0;
    }


  Matrix diag(const Matrix& Mat)
    {
      Tracer tr("diag");
      if(Mat.Ncols()==1){
	Matrix retmat(Mat.Nrows(),Mat.Nrows());
	diag(retmat,Mat);
	return retmat;}
      else{
	int mindim = Min(Mat.Ncols(),Mat.Nrows());
	Matrix retmat(mindim,1);
	for(int ctr=1; ctr<=mindim;ctr++){
	  retmat(ctr,1)=Mat(ctr,ctr);
	}
	return retmat;
      }
    }

  Matrix pinv(const Matrix& mat)
    {
      // calculates the psuedo-inverse using SVD
      Tracer tr("pinv");
      DiagonalMatrix D;
      Matrix U, V;
      SVD(mat,D,U,V);
      float tol;
      tol = MaximumAbsoluteValue(D) * Max(mat.Nrows(),mat.Ncols()) * 1e-16;
      for (int n=1; n<=D.Nrows(); n++) {
	if (fabs(D(n,n))>tol)  D(n,n) = 1.0/D(n,n);
      }
      Matrix pinv = V * D * U.t();
      return pinv;
    }


//    Matrix sqrtaff(const Matrix& mat)
//    {
//      complexmat lm;
//      lm = logm(mat);
//      lm /= 2.0;
//      Matrix sm;
//      sm = real(expm(lm));
//      sm = sm*sm - mat;
//      if (MaximumAbsoluteValue(sm)>1e-8) 
//        { cerr << "WARNING:: Sqrtaff badly conditioned result" << endl; }
//      return (real(exp(lm)));
//    }


  Matrix sqrtaff(const Matrix& mat)
    {
      Tracer tr("sqrtaff");
      Matrix matnew(4,4), rot(4,4), id4(4,4);
      Identity(rot);
      Identity(id4);
      ColumnVector params(12), centre(3), trans(4);
      centre = 0.0;
      // Quaternion decomposition -> params(1..3) = sin(theta/2)*(unit_axis_vec)
      // Want a new quaternion : q = sin(theta/4)*(unit_axis_vec)
      // Therefore factor of conversion is: factor = sin(theta/4)/sin(theta/2)
      //      = 1/(2  * cos(theta/4))   which is calculated below
      //  NB: t = theta/2
      decompose_aff(params,mat,centre,rotmat2quat);
      double sint;
      sint = std::sqrt(params(1)*params(1) + params(2)*params(2) + 
		  params(3)*params(3));
      double t = asin(sint);
      double factor = 1.0/(2.0*cos(0.5*t));
      params(1) = factor * params(1);
      params(2) = factor * params(2);
      params(3) = factor * params(3);
      params(7) = std::sqrt(params(7));
      params(8) = std::sqrt(params(8));
      params(9) = std::sqrt(params(9));
      params(10) = 0.5*params(10);
      params(11) = 0.5*params(11);
      params(12) = 0.5*params(12);

      construct_rotmat_quat(params,3,rot,centre);
      rot(1,4) = 0.0;
      rot(2,4) = 0.0;
      rot(3,4) = 0.0;
      
      Matrix scale(4,4);
      Identity(scale);
      scale(1,1)=params(7);
      scale(2,2)=params(8);
      scale(3,3)=params(9);

      Matrix skew(4,4);
      Identity(skew);
      skew(1,2)=params(10);
      skew(1,3)=params(11);
      skew(2,3)=params(12);

      trans(1) = params(4);
      trans(2) = params(5);
      trans(3) = params(6);
      trans(4) = 1.0;

      // The translation, being independent of the 3x3 submatrix, is
      //  calculated so that it will be equal for each of the two 
      //  halves of the approximate square root 
      //  (i.e. matnew and mat*matnew.i() have exactly the same translation)
      ColumnVector th(4);
      th = (mat*scale.i()*skew.i()*rot.i() + id4).SubMatrix(1,3,1,3).i() 
	* trans.SubMatrix(1,3,1,1);

      matnew = rot*skew*scale;
      matnew(1,4) = th(1);
      matnew(2,4) = th(2);
      matnew(3,4) = th(3);
      return matnew;
    }


  //------------------------------------------------------------------------//

  // Handy MATLAB-like functions
  
  void reshape(Matrix& r, const Matrix& m, int nrows, int ncols)
    {
      Tracer tr("reshape");
      if (nrows*ncols != m.Nrows() * m.Ncols() ) {
	cerr << "WARNING: cannot reshape " << m.Nrows() << "x"
	     << m.Ncols() << " matrix into " << nrows << "x" 
	     << ncols << endl;
	cerr << " Returning original matrix instead" << endl;
	r = m;
	return;
      }
      r.ReSize(nrows,ncols);
      int rr = 1, rc = 1;
      for (int mc=1; mc<=m.Ncols(); mc++) {
	for (int mr=1; mr<=m.Nrows(); mr++) {
	  r(rr,rc) = m(mr,mc);
	  rr++;
	  if (rr>nrows) {
	    rc++;
	    rr=1;
	  }
	}
      }
    }

  //------------------------------------------------------------------------//

  // Matrix routines, specialised for 4x4 matrices

   void matfuncore(const complexmat& m, complexmat& lambda, 
		   ColumnVector& mult,
		   veccomplexmat& beta, veccomplexmat& alpha,
		   veccomplexmat& gamma, veccomplexmat& delta) 
   {
     // Does the basic decomposition of a 4x4 affine to allow a function
     //  of m to be calculated
     // Note that m(4,1:3) = [0 0 0] for this to work, 
     //  but m(4,4) can be non-unity

     // Useful constants
     double eqtol = 1e-5;
     complex j(0,1);

     // Calculate the coefficients of the characteristic equation

     // 3x3 submatrix m(1:3,1:3)
     complex c1,c2,c3;
     c1 = -(m(1,1) + m(2,2) + m(3,3));
     c2 = m(1,1)*m(2,2) + m(1,1)*m(3,3) + m(2,2)*m(3,3) - m(1,2)*m(2,1)
       - m(3,2)*m(2,3) - m(1,3)*m(3,1);
     c3 = m(1,2)*m(2,1)*m(3,3) + m(1,3)*m(3,1)*m(2,2) + m(3,2)*m(2,3)*m(1,1)
       - m(1,1)*m(2,2)*m(3,3) - m(1,2)*m(2,3)*m(3,1) - m(1,3)*m(3,2)*
       m(2,1);

     // 4x4 characteristic equation coefficients
     complex l4, k1, k2, k3, k4;
     l4 = m(4,4);  // the 4th eigenvalue
     
     k1 = c1 - l4;
     k2 = c2 - l4*c1;
     k3 = c3 - l4*c2;
     k4 = -l4*c3;

     // Check characteristic coefficients
     //disp(poly(m) - [1 k1 k2 k3 k4])
     //disp([1 c1 c2 c3]);

     // Find roots of the cubic polynomial
     complex q,r,d,s,t,l1,l2,l3;
     q = (3.0*c2 - c1*c1)/9.0;
     r = (9.0*c1*c2 - 27.0*c3 - 2.0*(c1*c1*c1))/54.0;
     d = q*q*q + r*r;
     s = power(r + sqrt(d),(1.0/3.0));
     if (abs(s)>eqtol) {
       t = -q/s;
     } else {
       t = power(r - sqrt(d),(1.0/3.0));
     }
     l1 = s + t -c1/3.0;
     l2 = -(s+t)/2.0 -c1/3.0 + j*std::sqrt(3.0)*(s-t)/2.0;
     l3 = -(s+t)/2.0 -c1/3.0 - j*std::sqrt(3.0)*(s-t)/2.0;
     lambda.ReSize(4,1);
     lambda.set(1,1,l1);
     lambda.set(2,1,l2);
     lambda.set(3,1,l3);
     lambda.set(4,1,l4);

     // check eigenvalues:
     //[cv,cd]=eig(m);
     //disp(sort(diag(cd).'))
     //disp(sort(lambda))

     // determine multiplicity of the eigenvalues
     mult.ReSize(4);
     mult = 0;
     int sum;
     complex csum;
     for (int n=4; n>=1; n--) {
       sum=0;  csum=0.0;
       for (int k=4; k>=1; k--) { 
	 if ((k!=n) && (abs(lambda(k,1) - lambda(n,1))<max(eqtol,eqtol*abs(lambda(n,1)))))
	   {
	     sum++;
	     mult(k)=0;
	     csum += lambda(k,1);
	   }
       }
       mult(n) = sum+1;
       csum += lambda(n,1);
       csum /= mult(n);
       for (int k=4; k>=1; k--) { 
	 if (abs(lambda(k,1) - lambda(n,1))<max(eqtol,eqtol*abs(lambda(n,1)))) {
	   if (k!=n) lambda.set(k,1,csum);
	 }
       }
       lambda.set(n,1,csum);
     }
     //int msum=0;
     //for (int n=1; n<=4; n++)  msum+=(int) (mult(n)+0.5);
     //assert(msum==4);
     //disp([lambda mult]);

     // set up the initial recurrence parameters
     complexmat w0(Identity(4),Identity(4)*0.0);
     complexmat w1, w2, w3, x0, x1, x2, x3;
     w1 = m;
     w2 = m*m;
     w3 = m*m*m;
     
     x0 = w0;
     x1 = w1 + k1*w0;
     x2 = w2 + k1*w1 + k2*w0;
     x3 = w3 + k1*w2 + k2*w1 + k3*w0;
     
     // check Cayley-Hamilton
     //complexmat w4, rem;
     //w4 = m*m*m*m;
     //rem = w4 + k1*w3 + k2*w2 + k3*w1 + k4*w0;
     //print(rem);

     // initialize the parameters:
     //  beta, alpha, gamma, delta

     complexmat zeromat(4,4);
     zeromat = 0;
     beta.erase(beta.begin(),beta.end());
     alpha.erase(beta.begin(),beta.end());
     gamma.erase(beta.begin(),beta.end());
     delta.erase(beta.begin(),beta.end());
     for (int k=1; k<=4; k++) {
       beta.push_back(zeromat);
       alpha.push_back(zeromat);
       gamma.push_back(zeromat);
       delta.push_back(zeromat);
     }
       
     complexmat numpoly(4,4), numpoly1(4,4), numpoly2(4,4), numpoly3(4,4);
     complex denpoly;

     for (int n=1; n<=4; n++) {
       if (mult(n)>0) {
	 numpoly = x0*power(lambda(n,1),3) + x1*power(lambda(n,1),2)
	   + x2*lambda(n,1) + x3;
	 numpoly1 = 3*x0*power(lambda(n,1),2) + 2*x1*lambda(n,1) + x2;
	 numpoly2 = 6*x0*lambda(n,1) + 2*x1;
	 numpoly3 = 6*x0;
	 denpoly = 1;
	 for (int k=1; k<=4; k++) {
	   if ((k!=n) && (mult(k)>0)) {
	     denpoly = denpoly * power(lambda(n,1) - lambda(k,1),mult(k));
	   }
	 }
	 if (fabs(mult(n)-1.0)<0.5) {
	   beta[n-1] = numpoly / denpoly;
	 } else if (fabs(mult(n)-2.0)<0.5) {
	   alpha[n-1] = numpoly / denpoly;
	   beta[n-1]  = numpoly1 / denpoly;
	   for (int k=1; k<=4; k++) {
	     if ((k!=n) && (mult(k)>0)) {
	       beta[n-1] = beta[n-1] - mult(k)*numpoly/denpoly/(lambda(n,1) - lambda(k,1));
	     }
	   }
	 } else if (fabs(mult(n)-3.0)<0.5) {
	   gamma[n-1] = numpoly / denpoly;
	   alpha[n-1] = numpoly1 / denpoly;
	   for (int k=1; k<=4; k++) {
	     if ((k!=n) && (mult(k)>0)) {
	       alpha[n-1] = alpha[n-1] - mult(k)*numpoly/denpoly/(lambda(n,1)  - lambda(k,1));
	     }
	   }
	   // Note: the following is for the special case of a 4x4 matrix
	   beta[n-1] = (1.0/2.0)*(numpoly2/denpoly - 2.0*numpoly1/power(denpoly,2) 
			      + 2.0 * numpoly/power(denpoly,3));
	 } else if (fabs(mult(n)-4.0)<0.5) {
	   // Note: this is only for the special 4x4 case
	   delta[n-1] = numpoly;
	   gamma[n-1] = numpoly1;
	   alpha[n-1] = (1.0/2.0)*numpoly2;
	   beta[n-1]  = (1.0/6.0)*numpoly3;
	 }
       }
     }
   }
   


   complexmat expm(const complexmat& m) 
   {
     complexmat b, lambda;
     ColumnVector mult;
     veccomplexmat alpha, beta, gamma, delta;
     b = m * 0.0;
     matfuncore(m,lambda,mult,beta,alpha,gamma,delta);
     for (int n=1; n<=4; n++) {
       if (mult(n)>0) {
	 b += exp(lambda(n,1)) * beta[n-1];
         b += exp(lambda(n,1)) * alpha[n-1];
         b += (1.0/2.0)*exp(lambda(n,1)) * gamma[n-1];
         b += (1.0/6.0)*exp(lambda(n,1)) * delta[n-1];
       }
     }
     // check if original matrix was real (in which case the solution is too)
     Matrix ratio;
     ratio = imag(m);
     for (int n=1; n<=ratio.Nrows(); n++) {
       for (int k=1; k<=ratio.Ncols(); k++) {
	 ratio(n,k) /= m.real(n,k);
       }
     }
     if (ratio.MaximumAbsoluteValue()<1e-5) {
       b = real(b);
     }
     return b;
   }
   



   complexmat logm(const complexmat& m) 
   {
     complexmat b, lambda;
     ColumnVector mult;
     veccomplexmat alpha, beta, gamma, delta;
     b = m * 0.0;
     matfuncore(m,lambda,mult,beta,alpha,gamma,delta);
     for (int n=1; n<=4; n++) {
       if (mult(n)>0) {
	 b += log(lambda(n,1)) * beta[n-1];
         b += (1.0/lambda(n,1)) * alpha[n-1];
         b += (1.0/2.0)*(-1.0/power(lambda(n,1),2.0)) * gamma[n-1];
         b += (1.0/6.0)*(2.0/power(lambda(n,1),3.0)) * delta[n-1];
       }
     }
     return b;
   }
   

//     complexmat sqrtm(const complexmat& m) 
//     {
//       complexmat b, lambda;
//       ColumnVector mult;
//       veccomplexmat alpha, beta, gamma, delta;
//       b = m * 0.0;
//       matfuncore(m,lambda,mult,beta,alpha,gamma,delta);
//       for (int n=1; n<=4; n++) {
//         if (mult(n)>0) {
//  	 b += sqrt(lambda(n,1)) * beta[n-1];
//           b += (1.0/(2.0*power(lambda(n,1),1.0/2.0))) * alpha[n-1];
//           b += (1.0/2.0)*(-1.0/(4.0*power(lambda(n,1),3.0/2.0))) * gamma[n-1];
//           b += (1.0/6.0)*(3.0/(8.0*power(lambda(n,1),5.0/2.0))) * delta[n-1];
//         }
//       }
//       return b;
//     }
   


   Matrix expm(const Matrix& m) {
     complexmat res;
     res = expm(complexify(m));
     return real(res);
   }

   complexmat logm(const Matrix& m) { return logm(complexify(m)); }


  //------------------------------------------------------------------------//


  // Spatial transformation functions (rotations and affine transforms)


  int construct_rotmat_euler(const ColumnVector& params, int n, Matrix& aff,
			     const ColumnVector& centre)
    {
      Tracer tr("construct_rotmat_euler");
      ColumnVector angl(3);
      Matrix newaff(4,4);
      Identity(aff);

      if (n<=0) return 0;
      // order of parameters is 3 rotation + 3 translation
      // angles are in radians 
      //  order of parameters is (Rx,Ry,Rz) and R = Rx.Ry.Rz
      angl=0.0;
      angl(1)=params(1);
      make_rot(angl,centre,newaff);
      aff = aff * newaff;
      if (n==1) return 0;

      angl=0.0;
      angl(2)=params(2);
      make_rot(angl,centre,newaff);
      aff = aff * newaff; 
      if (n==2) return 0;

      angl=0.0;
      angl(3)=params(3);
      make_rot(angl,centre,newaff);
      aff = aff * newaff;
      if (n==3) return 0;

      aff(1,4)+=params(4);
      if (n==4) return 0;
      aff(2,4)+=params(5);
      if (n==5) return 0;
      aff(3,4)+=params(6);
      if (n==6) return 0;

      return 1;
    }  

  int construct_rotmat_euler(const ColumnVector& params, int n, Matrix& aff)
    {
      Tracer tr("construct_rotmat_euler");
      ColumnVector centre(3);
      centre = 0.0;
      return construct_rotmat_euler(params,n,aff,centre);
    }


  int construct_rotmat_quat(const ColumnVector& params, int n, Matrix& aff,
			    const ColumnVector& centre)
    {
      Tracer tr("construct_rotmat_quat");
      Identity(aff);

      if (n<=0) return 0;
      // order of parameters is 3 rotation (last 3 quaternion components) 
      //  + 3 translation

      if ((n>=1) && (n<3)) { cerr<<"Can only do 3 or more, not "<< n <<endl; }
      float w, w2 = 1.0 - Sqr(params(1)) - Sqr(params(2)) - Sqr(params(3));
      if (w2 < 0.0) {
	cerr << "Parameters do not form a valid axis - greater than unity\n";
	return -1;
      }
      w = std::sqrt(w2);
      float x=params(1), y=params(2), z=params(3);
      aff(1,1) = 1 - 2*y*y - 2*z*z;
      aff(2,2) = 1 - 2*x*x - 2*z*z;
      aff(3,3) = 1 - 2*x*x - 2*y*y;
      aff(1,2) = 2*x*y - 2*w*z;
      aff(2,1) = 2*x*y + 2*w*z;
      aff(1,3) = 2*x*z + 2*w*y;
      aff(3,1) = 2*x*z - 2*w*y;
      aff(2,3) = 2*y*z - 2*w*x;
      aff(3,2) = 2*y*z + 2*w*x;

      // Given Rotation matrix R:  x' = Rx + (I-R)*centre
      ColumnVector trans(3);
      trans = aff.SubMatrix(1,3,1,3)*centre;
      aff.SubMatrix(1,3,4,4) = centre - trans;

      aff(1,4) += params(4);
      if (n==4) return 0;
      aff(2,4) += params(5);
      if (n==5) return 0;
      aff(3,4) += params(6);
      if (n==6) return 0;

      return 1;
    }  

  int construct_rotmat_quat(const ColumnVector& params, int n, Matrix& aff)
    {
      Tracer tr("construct_rotmat_quat");
      ColumnVector centre(3);
      centre = 0.0;
      return construct_rotmat_quat(params,n,aff,centre);
    }

  int make_rot(const ColumnVector& angl, const ColumnVector& centre, 
	       Matrix& rot)
    {
      // Matrix rot must be 4x4; angl and orig must be length 3
      Tracer tr("make_rot");
      Identity(rot);  // default return value
      float theta;
      theta = norm2(angl);
      if (theta<1e-8) {  // avoid round-off errors and return Identity
	return 0;
      }
      ColumnVector axis = angl/theta;
      ColumnVector x1(3), x2(3), x3(3);
      x1 = axis;
      x2(1) = -axis(2);  x2(2) = axis(1);  x2(3) = 0.0;
      if (norm2(x2)<=0.0) {
	x2(1) = 1.0;  x2(2) = 0.0;  x2(3) = 0.0;
      }
      x2 = x2/norm2(x2);
      x3 = cross(x1,x2);
      x3 = x3/norm2(x3);

      Matrix basischange(3,3);
      basischange.SubMatrix(1,3,1,1) = x2;
      basischange.SubMatrix(1,3,2,2) = x3;
      basischange.SubMatrix(1,3,3,3) = x1;

      Matrix rotcore(3,3);
      Identity(rotcore);
      rotcore(1,1)=cos(theta);
      rotcore(2,2)=cos(theta);
      rotcore(1,2)=sin(theta);
      rotcore(2,1)=-sin(theta);

      rot.SubMatrix(1,3,1,3) = basischange * rotcore * basischange.t();
  
      Matrix ident3(3,3);
      Identity(ident3);
      ColumnVector trans(3);
      trans = (ident3 - rot.SubMatrix(1,3,1,3))*centre;
      rot.SubMatrix(1,3,4,4)=trans;
      return 0;
    }


  int getrotaxis(ColumnVector& axis, const Matrix& rotmat)
    {
      Tracer tr("getrotaxis");
      Matrix residuals(3,3);
      residuals = rotmat*rotmat.t() - Identity(3);
      if (residuals.SumSquare() > 1e-4)
	{ cerr << "Failed orthogonality check!" << endl;  return -1; }
      Matrix u(3,3), v(3,3);
      DiagonalMatrix d(3);
      SVD(rotmat-Identity(3),d,u,v);
      // return column of V corresponding to minimum value of |S|
      for (int i=1; i<=3; i++) {
	if (fabs(d(i))<1e-4)  axis = v.SubMatrix(1,3,i,i);
      }
      return 0;
    }  
  

  int rotmat2euler(ColumnVector& angles, const Matrix& rotmat)
    {
      // uses the convention that R = Rx.Ry.Rz
      Tracer tr("rotmat2euler");
      float cz, sz, cy, sy, cx, sx;
      cy = std::sqrt(Sqr(rotmat(1,1)) + Sqr(rotmat(1,2)));
      if (cy < 1e-4) {
	//cerr << "Cos y is too small - Gimbal lock condition..." << endl;
	cx = rotmat(2,2);
	sx = -rotmat(3,2);
	sy = -rotmat(1,3);
	angles(1) = atan2(sx,cx);
	angles(2) = atan2(sy,0.0);
	angles(3) = 0.0;
      } else {
	// choose by convention that cy > 0
	// get the same rotation if: sy stays same & all other values swap sign
	cz = rotmat(1,1)/cy;
	sz = rotmat(1,2)/cy;
	cx = rotmat(3,3)/cy;
	sx = rotmat(2,3)/cy;
	sy = -rotmat(1,3);
	//atan2(sin,cos)  (defined as atan2(y,x))
	angles(1) = atan2(sx,cx);
	angles(2) = atan2(sy,cy);
	angles(3) = atan2(sz,cz);
      }
      return 0;
    }


  int rotmat2quat(ColumnVector& quaternion, const Matrix& rotmat)
    {
      Tracer tr("rotmat2quat");
      float w = std::sqrt((rotmat.SubMatrix(1,3,1,3).Trace() + 1.0)/4.0);
      quaternion(1) = (rotmat(3,2) - rotmat(2,3))/(4.0*w);
      quaternion(2) = (rotmat(1,3) - rotmat(3,1))/(4.0*w);
      quaternion(3) = (rotmat(2,1) - rotmat(1,2))/(4.0*w);
      return 0;
    }


  int decompose_aff(ColumnVector& params, const Matrix& affmat, 
		    const ColumnVector& centre,
		    int (*rotmat2params)(ColumnVector& , const Matrix& ))
    {
      // decomposes using the convention: mat = rotmat * skew * scale
      // order of parameters is 3 rotation + 3 translation + 3 scales + 3 skews
      // angles are in radians
      Tracer tr("decompose_aff");
      if (params. Nrows() < 12)
	params.ReSize(12);
      if (rotmat2params==0)  
	{ 
	  cerr << "No rotmat2params function specified" << endl;  
	  return -1; 
	}
      ColumnVector x(3), y(3), z(3);
      Matrix aff3(3,3);
      aff3 = affmat.SubMatrix(1,3,1,3);
      x = affmat.SubMatrix(1,3,1,1);
      y = affmat.SubMatrix(1,3,2,2);
      z = affmat.SubMatrix(1,3,3,3);
      float sx, sy, sz, a, b, c;
      sx = norm2(x);
      sy = std::sqrt( dot(y,y) - (Sqr(dot(x,y)) / Sqr(sx)) );
      a = dot(x,y)/(sx*sy);
      ColumnVector x0(3), y0(3);
      x0 = x/sx;
      y0 = y/sy - a*x0;
      sz = std::sqrt(dot(z,z) - Sqr(dot(x0,z)) - Sqr(dot(y0,z)));
      b = dot(x0,z)/sz;
      c = dot(y0,z)/sz;
      params(7) = sx;  params(8) = sy;  params(9) = sz;
      Matrix scales(3,3);
      float diagvals[] = {sx,sy,sz};
      diag(scales,diagvals);
      Real skewvals[] = {1,a,b,0 , 0,1,c,0 , 0,0,1,0 , 0,0,0,1}; 
      Matrix skew(4,4);
      skew  << skewvals;
      params(10) = a;  params(11) = b;  params(12) = c;
      Matrix rotmat(3,3);
      rotmat = aff3 * scales.i() * (skew.SubMatrix(1,3,1,3)).i();
      ColumnVector transl(3);
      //transl = affmat.SubMatrix(1,3,4,4);
      //transl = transl - (Identity(3) - rotmat)*centre;
      transl = affmat.SubMatrix(1,3,1,3)*centre + affmat.SubMatrix(1,3,4,4)
	         - centre;
      for (int i=1; i<=3; i++)  { params(i+3) = transl(i); }
      ColumnVector rotparams(3);
      (*rotmat2params)(rotparams,rotmat);
      for (int i=1; i<=3; i++)  { params(i) = rotparams(i); }
      return 0;
    }

  int decompose_aff(ColumnVector& params, const Matrix& affmat, 
		    int (*rotmat2params)(ColumnVector& , const Matrix& ))
    {
      Tracer tr("decompose_aff");
      ColumnVector centre(3);
      centre = 0.0;
      return decompose_aff(params,affmat,centre,rotmat2params);
    }



  int compose_aff(const ColumnVector& params, int n, const ColumnVector& centre,
		  Matrix& aff, 
		  int (*params2rotmat)(const ColumnVector& , int , Matrix& ,
			     const ColumnVector& ) )
    {
      Tracer tr("compose_aff");
      if (n<=0) return 0;
      // order of parameters is 3 rotation + 3 translation + 3 scales + 3 skews
      // angles are in radians

      (*params2rotmat)(params,n,aff,centre);
  
      if (n<=6)  return 0;
  
      Matrix scale(4,4);
      Identity(scale);
      if (n>=7) {
	scale(1,1)=params(7);
	if (n>=8) scale(2,2)=params(8);
	else      scale(2,2)=params(7);
	if (n>=9) scale(3,3)=params(9);
	else      scale(3,3)=params(7);
      }
      // fix the translation so that the centre is not moved
      ColumnVector strans(3);
      strans = centre - scale.SubMatrix(1,3,1,3)*centre;
      scale.SubMatrix(1,3,4,4) = strans;

      Matrix skew(4,4);
      Identity(skew);
      if (n>=10) {
	if (n>=10) skew(1,2)=params(10);
	if (n>=11) skew(1,3)=params(11);
	if (n>=12) skew(2,3)=params(12);
      }
      // fix the translation so that the centre is not moved
      ColumnVector ktrans(3);
      ktrans = centre - skew.SubMatrix(1,3,1,3)*centre;
      skew.SubMatrix(1,3,4,4) = ktrans;

      aff = aff * skew * scale;

      return 0;
    }


float rms_deviation(const Matrix& affmat1, const Matrix& affmat2, 
		    const ColumnVector& centre, const float rmax) 
{
  Tracer trcr("rms_deviation");
  Matrix isodiff(4,4);
  try {
    isodiff = affmat1*affmat2.i() - Identity(4);
  } catch(...) {
    cerr << "RMS_DEVIATION ERROR:: Could not invert matrix" << endl;  
    exit(-5); 
  }
  Matrix adiff(3,3);
  adiff = isodiff.SubMatrix(1,3,1,3);
  ColumnVector tr(3);
  tr = isodiff.SubMatrix(1,3,4,4) + adiff*centre;
  float rms = std::sqrt( (tr.t() * tr).AsScalar() + 
		    (rmax*rmax/5.0)*Trace(adiff.t()*adiff) );
  return rms;
}


float rms_deviation(const Matrix& affmat1, const Matrix& affmat2, 
		    const float rmax) 
{
  ColumnVector centre(3);
  centre = 0;
  return rms_deviation(affmat1,affmat2,centre,rmax);
}


// Added by MWW

//  int getdiag(ColumnVector& diagvals, const Matrix& m)
//  {
//    Tracer ts("MiscMaths::diag"); 
      
//    int num = m.Nrows();
//    diagvals.ReSize(num);
//    for (int j=1; j<=num; j++)
//      diagvals(j)=m(j,j);
//    return 0;
//  }

float var(const ColumnVector& x)
{     
  float m = mean(x);
  float ssq = (x-m).SumSquare()/(x.Nrows()-1);
      
  return ssq;
}

float mean(const ColumnVector& x)
{    
  float m = x.Sum()/x.Nrows();
  return m;
}

float median(const ColumnVector& x)
{     
  ColumnVector y = x;
  SortAscending(y);
  float m = y((int)(y.Nrows()/2));
      
  return m;
}

// Added by CFB   --- Matlab style Matrix functions
 
Matrix ones(const int dim1, const int dim2)
{ 
  int tdim = dim2;
  if(tdim<0){tdim=dim1;}
  Matrix res(dim1,tdim); res = 1.0;
  return res;
}

Matrix zeros(const int dim1, const int dim2)
{ 
  int tdim = dim2;
  if(tdim<0){tdim=dim1;}
  Matrix res(dim1,tdim); res = 0.0;
  return res;
}

Matrix repmat(const Matrix &mat, const int rows, const int cols)
{
  Matrix res = mat;
  for(int ctr = 1; ctr < cols; ctr++){res |= mat;}
  Matrix tmpres = res;
  for(int ctr = 1; ctr < rows; ctr++){res &= tmpres;}
  return res;
}

Matrix dist2(const Matrix &mat1, const Matrix &mat2)
{
  Matrix res(mat1.Ncols(),mat2.Ncols());
  for(int ctr1 = 1; ctr1 <= mat1.Ncols(); ctr1++)
    for(int ctr2 =1; ctr2 <= mat2.Ncols(); ctr2++)
      {
	ColumnVector tmp;
	tmp=mat1.Column(ctr1)-mat2.Column(ctr2);
	res(ctr1,ctr2) = std::sqrt(tmp.SumSquare());
      }
  return res;
}

Matrix abs(const Matrix& mat)
{
  Matrix res = mat;
  for (int mc=1; mc<=mat.Ncols(); mc++) {
    for (int mr=1; mr<=mat.Nrows(); mr++) {
      res(mr,mc)=std::abs(res(mr,mc));
    }
  }
  return res;
}

Matrix sqrt(const Matrix& mat)
{
  Matrix res = mat;
  bool neg_flag = false;
  for (int mc=1; mc<=mat.Ncols(); mc++) {
    for (int mr=1; mr<=mat.Nrows(); mr++) {
      if(res(mr,mc)<0){ neg_flag = true; }
      res(mr,mc)=std::sqrt(std::abs(res(mr,mc)));
    }
  }
  if(neg_flag){
    cerr << " Matrix contained negative elements" << endl;
    cerr << " return sqrt(abs(X)) instead" << endl;
  }
  return res;
}

Matrix log(const Matrix& mat)
{
  Matrix res = mat;
  bool neg_flag = false;
  for (int mc=1; mc<=mat.Ncols(); mc++) {
    for (int mr=1; mr<=mat.Nrows(); mr++) {
      if(res(mr,mc)<0){ neg_flag = true; }
      res(mr,mc)=std::log(std::abs(res(mr,mc)));
    }
  }
  if(neg_flag){
    cerr << " Matrix contained negative elements" << endl;
    cerr << " return log(abs(X)) instead" << endl;
  }
  return res; 
}

Matrix exp(const Matrix& mat)
{
  Matrix res = mat;
  for (int mc=1; mc<=mat.Ncols(); mc++) {
    for (int mr=1; mr<=mat.Nrows(); mr++) {
      res(mr,mc)=std::exp(res(mr,mc));
    }
  }
  return res;
}

Matrix tanh(const Matrix& mat)
{
  Matrix res = mat;
  for (int mc=1; mc<=mat.Ncols(); mc++) {
    for (int mr=1; mr<=mat.Nrows(); mr++) {
      res(mr,mc)=std::tanh(res(mr,mc));
    }
  }
  return res;
}

Matrix pow(const Matrix& mat, const double exp)
{
  Matrix res = mat;
  for (int mc=1; mc<=mat.Ncols(); mc++) {
    for (int mr=1; mr<=mat.Nrows(); mr++) {
      res(mr,mc)=std::pow(res(mr,mc),exp);
    }
  }
  return res;
}

Matrix sum(const Matrix& mat, const int dim)
{
  Matrix tmp;

  if (dim == 1) {tmp=mat;}
  else {tmp=mat.t();}
  Matrix res(1,tmp.Ncols());
  res = 0.0;  
  for (int mc=1; mc<=tmp.Ncols(); mc++) {
    for (int mr=1; mr<=tmp.Nrows(); mr++) {
      res(1,mc) += tmp(mr,mc);
    }
  }
  if (dim == 1) {return res;}
  else {return res.t();}
}

Matrix mean(const Matrix& mat, const int dim)
{
  Matrix res;
  res = sum(mat,dim);
  int N;
  if (dim == 1){N = mat.Nrows();}else{N = mat.Ncols();}
  return res/N;
}

Matrix var(const Matrix& mat, const int dim)
{
  Matrix res;
  Matrix res2;
  int N;
  res = pow(sum(mat,dim),2);
  res2= sum(pow(mat,2),dim);
  if (dim == 1){N = mat.Nrows();}else{N = mat.Ncols();}
  return res2/(N-1)-res/(N*(N-1));
}

Matrix stdev(const Matrix& mat, const int dim)
{
  return sqrt(var(mat,dim));
}

Matrix remmean(const Matrix& mat, const int dim)
{ 
  Matrix res;
  if (dim == 1) {res=mat;}
  else {res=mat.t();}

  Matrix Mean;
  Mean = mean(res);

  for (int ctr = 1; ctr <= res.Nrows(); ctr++) {
    for (int ctr2 =1; ctr2 <= res.Ncols(); ctr2++) {
      res(ctr,ctr2)-=Mean(1,ctr2);
    }
  }
  if (dim == 1) {return res;}
  else {return res.t();}
}

SymmetricMatrix cov(const Matrix& mat, const int norm)
{ 
  SymmetricMatrix res;
  Matrix tmp;
  int N;
  tmp=remmean(mat);
  if (norm == 1) {N = mat.Nrows();}
  else {N = mat.Nrows()-1;}  
  res << tmp.t()*tmp;

  return res/N; 
}

SymmetricMatrix corrcoef(const Matrix& mat, const int norm)
{ 
  SymmetricMatrix res;
  SymmetricMatrix C;
  C = cov(mat,norm);
  Matrix D;
  D=diag(C);
  D=pow(sqrt(D*D.t()),-1);
  res << SP(C,D);
  return res; 
}

void symm_orth(Matrix &Mat)
{
  SymmetricMatrix Metric;
  Metric << Mat.t()*Mat;
  Metric = Metric.i();
  Matrix tmpE;
  DiagonalMatrix tmpD;
  EigenValues(Metric,tmpD,tmpE);
  Mat = Mat * tmpE * sqrt(abs(tmpD)) * tmpE.t();
}

void powerspectrum(const Matrix &Mat1, Matrix &Result, bool useLog)
  //calculates the powerspectrum for every column of Mat1
{
  Matrix res;
  for(int ctr=1; ctr <= Mat1.Ncols(); ctr++)
    {
      ColumnVector tmpCol;
      tmpCol=Mat1.Column(ctr);
      ColumnVector FtmpCol_real;
      ColumnVector FtmpCol_imag;
      ColumnVector tmpPow;
      
      RealFFT(tmpCol,FtmpCol_real,FtmpCol_imag);
      tmpPow = pow(FtmpCol_real,2)+pow(FtmpCol_imag,2);
      tmpPow = tmpPow.Rows(2,tmpPow.Nrows());
      if(useLog){tmpPow = log(tmpPow);}
      if(res.Storage()==0){res= tmpPow;}else{res|=tmpPow;}
    }
    Result=res;
}


int nextpow2(int n)
{
  return (int)pow(2,ceil(log(n)/log(2)));
}

void acorr(const Matrix& p_ts, Matrix& ret, int p_zeropad)
{
  Tracer tr("MISCMATHS::acorr");

  int sizeTS = p_ts.Nrows();
  int numTS = p_ts.Ncols();
  
  if(p_zeropad == 0)
    p_zeropad = sizeTS;

  ColumnVector x(p_zeropad);
  x = 0;
  ColumnVector fft_real;
  ColumnVector fft_im;
  ColumnVector dummy(p_zeropad);
  ColumnVector dummy2;
  dummy = 0;
  ColumnVector realifft(p_zeropad);
  ret.ReSize(sizeTS,numTS);
  
  for(int i = 1; i <= numTS; i++)
    {
      x.Rows(1,sizeTS) = p_ts.Column(i);
      FFT(x, dummy, fft_real, fft_im);
      
      for(int j = 1; j <= p_zeropad; j++)
	{
	  // (x+iy)(x-iy) = x^2 + y^2
	  fft_real(j) = fft_real(j)*fft_real(j) + fft_im(j)*fft_im(j);
	  fft_im(j) = 0;
	}
      
      FFTI(fft_real, fft_im, realifft, dummy2);
      
      float varx = var(x.Rows(1,sizeTS));
      ret.Column(i) = realifft.Rows(1,sizeTS);
      
      for(int j = 1; j <= sizeTS-1; j++)
	{
	  // Correction to make autocorr unbiased and normalised
	  ret(j,i) = ret(j,i)/((sizeTS-j)*varx);
	}  
    }
}

void detrend(Matrix& p_ts, int p_level)
{
  Tracer trace("MISCMATHS::detrend");

  int sizeTS = p_ts.Nrows();     

  // p_ts = b*a + e (OLS regression)
  // e is detrended data
  Matrix a(sizeTS, p_level+1);  

  // Create a
  for(int t = 1; t <= sizeTS; t++)
    {
      for(int l = 0; l <= p_level; l++)
	a(t,l+1) = pow((float)t/sizeTS,l);
    }
          
  // Form residual forming matrix R:
  Matrix R = Identity(sizeTS)-a*pinv(a);

  for(int t = 1; t <= sizeTS; t++)
    {
      p_ts.Column(t) = R*p_ts.Column(t);
    }      
}

// Complex Matrix class stuff

   // helper functions for the Matrix class
   Matrix real(const complexmat& mat) { return mat.real(); }
   Matrix imag(const complexmat& mat) { return mat.imag(); }
   complexmat conj(const complexmat& mat) { return mat.conj(); }
   Matrix Absolute(const complexmat& mat) { return mat.Absolute(); }
   complex Trace(const complexmat& mat) { return mat.Trace(); }
   complexmat SP(const complexmat& mat1, const complexmat& mat2)
     { 
      return  complexmat(SP(mat1.real(),mat2.real()) - SP(mat1.imag(),mat2.imag()),
		      SP(mat1.real(),mat2.imag()) + SP(mat1.imag(),mat2.real()));
     }
   complexmat complexify(const Matrix& mat)
     { return complexmat(mat,0.0*mat); }
   complex AsScalar(const complexmat& mat) { return mat.AsScalar(); }
   complex Sum(const complexmat& mat) { return mat.Sum(); }
   Real SumAbsoluteValue(const complexmat& mat) 
      { return mat.SumAbsoluteValue(); }
   bool IsZero(const complexmat& mat) { return mat.IsZero(); }

   complexmat operator+(const Real& v, const complexmat& mat) 
      { complexmat res(mat);  res += v;  return res; }
   complexmat operator-(const Real& v, const complexmat& mat) 
      { complexmat res(mat);  res -= v;  res *= -1;  return res; }
   complexmat operator*(const Real& v, const complexmat& mat) 
      { complexmat res(mat);  res *= v;  return res; }
   complexmat operator+(const complex& v, const complexmat& mat) 
      { complexmat res(mat);  res += v;  return res; }
   complexmat operator-(const complex& v, const complexmat& mat) 
      { complexmat res(mat);  res -= v;  res *= -1;  return res; }
   complexmat operator*(const complex& v, const complexmat& mat) 
      { complexmat res(mat);  res *= v;  return res; }

   void FFT(const complexmat& X, complexmat& F) {
     ColumnVector fr, fi;
     FFT(real(X).AsColumn(),imag(X).AsColumn(),fr,fi);
     F.setreal(fr.AsMatrix(fr.Nrows(),1));
     F.setimag(fi.AsMatrix(fr.Nrows(),1));
   }
   void FFTI(const complexmat& F, complexmat& X) {
     ColumnVector xr, xi;
     FFTI(real(F).AsColumn(),imag(F).AsColumn(),xr,xi);
     X.setreal(xr.AsMatrix(xr.Nrows(),1));
     X.setimag(xi.AsMatrix(xi.Nrows(),1));
   }

   void print(const complexmat& v) 
   {
     for (int x=1; x<=real(v).Nrows(); x++) {
       for (int y=1; y<=real(v).Ncols(); y++) {
	 cout << v.real(x,y) << " + " << v.imag(x,y) << "j	";
       }
       cout << endl;
     }
   }
}
