/*  interpolation.cc

    Mark Jenkinson, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

// Interpolation functions


#include "interpolation.h"
#include <iostream>
#include <math.h>

using namespace MISCMATHS;

#ifndef NO_NAMESPACE
 namespace INTERPOLATION {
#endif

  // INTERPOLATION STUFF

  float basic_tri_interpolation(const volume& v, 
				const float dx, const float dy, const float dz,
				const float v000, const float v001,
				const float v010, const float v011,
				const float v100, const float v101,
				const float v110, const float v111)
    {
      const float dx1=1.0-dx, dy1=1.0-dy, dz1=1.0-dz;
 

      return ( dx * ( dy * ( v111*dz  + 
			     v110*dz1 ) +
		      dy1 * ( v101*dz + 
			      v100*dz1 ) ) + 
	       dx1 * ( dy * ( v011*dz +
			      v010*dz1 )  +
		       dy1 * ( v001*dz  + 
			       v000*dz1 ) ) );
			   
    }

  
  float tri_interpolation(const volume& v, const float x, const float y,
			  const float z)
    {
      int ix0, iy0, iz0, ix1, iy1, iz1;
      ix0 = (int) floor(x);
      iy0 = (int) floor(y);
      iz0 = (int) floor(z);
      ix1 = ix0 + 1;
      iy1 = iy0 + 1;
      iz1 = iz0 + 1;
      if ((ix0==v.columns()-1) && ((x-(float)ix0)<1e-5))
	{
	  ix1=ix0;
	}
      if ((iy0==v.rows()-1) && ((y-(float)iy0)<1e-5))
	{
	  iy1=iy0;
	}
      if ((iz0==v.slices()-1) && ((z-(float)iz0)<1e-5))
	{
	  iz1=iz0;
	}
      if ( (!v.in_bounds(ix0,iy0,iz0)) || (!v.in_bounds(ix1,iy1,iz1)) )
	{
	  cerr << "Tri_interpolation: Out of bounds!\n";
	  cerr << "(" << x << "," << y << "," << z << ") vs ";
	  cerr << "(" << v.columns() << "," << v.rows() << "," << 
	    v.slices() << ")" << endl;
	}
      return basic_tri_interpolation(v,x-ix0,y-iy0,z-iz0,v(ix0,iy0,iz0),
				     v(ix0,iy0,iz1),v(ix0,iy1,iz0),
				     v(ix0,iy1,iz1),v(ix1,iy0,iz0),
				     v(ix1,iy0,iz1),v(ix1,iy1,iz0),
				     v(ix1,iy1,iz1));
    }


  bool filled_bounds(int x, int y, int z, int xb, int yb, int zb)
    {
      return ( (x>=0) && (y>=0) && (z>=0) && 
	       (x<=xb) && (y<=yb) && (z<=zb) );
    }

  

  float filled_tri_interpolation(const volume& v, const float x, const float y,
			  const float z)
    {
      int ix0, iy0, iz0, ix1, iy1, iz1;
      ix0 = (int) floor(x);
      iy0 = (int) floor(y);
      iz0 = (int) floor(z);
      float dx=x-ix0, dy=y-iy0, dz=z-iz0;
      ix1 = ix0 + 1;
      iy1 = iy0 + 1;
      iz1 = iz0 + 1;

      // extend the outer edges by one voxel

      if (ix0==v.xsize()-1) { ix1=ix0;  dx=0.0; }
      else if (ix0==-1) { ix0=0; ix1=ix0; dx=0.0; }

      if (iy0==v.ysize()-1) { iy1=iy0;  dy=0.0; }
      else if (iy0==-1) { iy0=0; iy1=iy0; dy=0.0; }

      if (iz0==v.zsize()-1) { iz1=iz0;  dz=0.0; }
      else if (iz0==-1) { iz0=0; iz1=iz0; dz=0.0; }


      if ( (!v.in_bounds(ix0,iy0,iz0)) || (!v.in_bounds(ix1,iy1,iz1)) )
	{
//  	  cerr << "Tri_interpolation: Out of bounds!\n";
//  	  cerr << "(" << x << "," << y << "," << z << ") vs ";
//  	  cerr << "(" << v.columns() << "," << v.rows() << "," << 
//  	    v.slices() << ")" << endl;
	  return v.backgroundval();
	}
      return basic_tri_interpolation(v,dx,dy,dz,
				     v(ix0,iy0,iz0),
				     v(ix0,iy0,iz1),v(ix0,iy1,iz0),
				     v(ix0,iy1,iz1),v(ix1,iy0,iz0),
				     v(ix1,iy0,iz1),v(ix1,iy1,iz0),
				     v(ix1,iy1,iz1));
    }



  float bi_interpolation(const volume& v, const float x, const float y)
    {
      int ix0, iy0, ix1, iy1;
      ix0 = (int) floor(x);
      iy0 = (int) floor(y);
      ix1 = ix0 + 1;
      iy1 = iy0 + 1;
      if ((ix0==v.columns()-1) && ((x-(float)ix0)<1e-5))
	{
	  ix1=ix0;
	}
      if ((iy0==v.rows()-1) && ((y-(float)iy0)<1e-5))
	{
	  iy1=iy0;
	}
      if ( (!v.in_bounds(ix0,iy0,0)) || (!v.in_bounds(ix1,iy1,0)) )
	{
	  cerr << "Bi_interpolation: Out of bounds!\n";
	  cerr << "(" << x << "," << y << ") vs ";
	  cerr << "(" << v.columns() << "," << v.rows() << ")" << endl;
	  return 0.0;
	}
      return basic_tri_interpolation(v,x-ix0,y-iy0,0.0,v(ix0,iy0,0),
				     v(ix0,iy0,0),v(ix0,iy1,0),
				     v(ix0,iy1,0),v(ix1,iy0,0),
				     v(ix1,iy0,0),v(ix1,iy1,0),
				     v(ix1,iy1,0));
    }


   ////////////////////////////////////////////////////////////////////////
   // OTHER INTERPOLATION METHODS (all filled versions)
   ////////////////////////////////////////////////////////////////////////

  float nn_interpolation(const volume& v, const float x, const float y,
			 const float z)
    {
      int ix0, iy0, iz0;
      ix0 = (int) floor(x + 0.5);
      iy0 = (int) floor(y + 0.5);
      iz0 = (int) floor(z + 0.5);
      // expand to be the filled version
      if (ix0==-1) ix0=0;
      if (iy0==-1) iy0=0;
      if (iz0==-1) iz0=0;
      if (ix0==v.xsize()) ix0--;
      if (iy0==v.ysize()) iy0--;
      if (iz0==v.zsize()) iz0--;
      if (!v.in_bounds(ix0,iy0,iz0))
	{
//  	  cerr << "NN_interpolation: Out of bounds!\n";
//  	  cerr << "(" << x << "," << y << "," << z << ") vs ";
//  	  cerr << "(" << v.columns() << "," << v.rows() << "," << 
//  	    v.slices() << ")" << endl;
	  return v.backgroundval();
	}
      return v(ix0,iy0,iz0);
    }

   ////////////////////////////////////////////////////////////////////////

   float sinc(float x)
     {
       if (fabs(x)<1e-7) { return 1.0-fabs(x); }
       float y=M_PI*x;
       return sin(y)/y;
     }

   float hanning(float x, int w)
     {
       if (fabs(x)>w) 
	 return 0.0;
       else
	 return (0.5 + 0.5 *cos(M_PI*x/w));
     }

   float blackman(float x, int w)
     {
       if (fabs(x)>w) 
	 return 0.0;
       else
	 return (0.42 + 0.5 *cos(M_PI*x/w) + 0.08*cos(2.0*M_PI*x/w));
     }

   float rectangular(float x, int w)
   {
     if (fabs(x)>w) 
       return 0.0;
     else
       return 1.0;
   }


   sincinterp::sincinterp()
   {
     kernelwidth_x=0;
     kernelwidth_y=0;
     kernelwidth_z=0;
     sincz = 0;
     sincy = 0;
     sincx = 0;
   }

   sincinterp::~sincinterp()
   {
     if (sincz) delete [] sincz;
     if (sincy) delete [] sincy;
     if (sincx) delete [] sincx;
   }

   void sincinterp::setupkernel(int wx, int wy, int wz, const string &sincwin)
     {
       kernelwidth_x=wx;
       kernelwidth_y=wy;
       kernelwidth_z=wz;
       if (sincz) delete [] sincz;
       if (sincy) delete [] sincy;
       if (sincx) delete [] sincx;
       sincz = new float[2*wz+1];
       sincy = new float[2*wy+1];
       sincx = new float[2*wx+1];
       // set x between +/- kernelwidthmax
       float halfnk = (NSINCK-1.0)/2.0;
       for (int n=0; n<NSINCK; n++) {
	 float x=(n-halfnk)/halfnk*kernelwidth_x;
	 float y=(n-halfnk)/halfnk*kernelwidth_y;
	 float z=(n-halfnk)/halfnk*kernelwidth_z;
	 if (sincwin=="hanning") {
	   sinckernelx[n] = sinc(x)*hanning(x,kernelwidth_x);
	   sinckernely[n] = sinc(y)*hanning(y,kernelwidth_y);
	   sinckernelz[n] = sinc(z)*hanning(z,kernelwidth_z);
	 } else if (sincwin=="blackman") {
	   sinckernelx[n] = sinc(x)*blackman(x,kernelwidth_x);
	   sinckernely[n] = sinc(y)*blackman(y,kernelwidth_y);
	   sinckernelz[n] = sinc(z)*blackman(z,kernelwidth_z);
	 } else if (sincwin=="rectangular") {
	   sinckernelx[n] = sinc(x)*rectangular(x,kernelwidth_x);
	   sinckernely[n] = sinc(y)*rectangular(y,kernelwidth_y);
	   sinckernelz[n] = sinc(z)*rectangular(z,kernelwidth_z);
	 }
       }
     }


   float sincinterp::kernelval(float x, int w, float *kernel)
     {
       // effectively returns  sinc(x)*hanning(x,w);
       if (fabs(x)>w) return 0.0;
       float halfnk = (NSINCK-1.0)/2.0;
       float dn = x/w*halfnk + halfnk;
       int n = (int) floor(dn);
       dn -= n;
       if (n>NSINCK) return 0.0;
       if (n<0) return 0.0;

       return kernel[n]*(1.0-dn) + kernel[n+1]*dn;
     }


   float sincinterp::sinc_interpolation(const volume& v, 
					const float x, const float y,
					const float z)
     {
          // kernel half-width  (i.e. range is +/- w)
       int wx=kernelwidth_x;  
       int wy=kernelwidth_y;  
       int wz=kernelwidth_z;  
       if ( (wx<3) || (wy<3) || (wz<3) ) { 
	 setupkernel(3,3,3,"hanning");  // the default is 3 and Hanning
	 wx=kernelwidth_x;
	 wy=kernelwidth_y;
	 wz=kernelwidth_z;
       }
       
       int ix0, iy0, iz0;
       ix0 = (int) floor(x);
       iy0 = (int) floor(y);
       iz0 = (int) floor(z);

       float convsum=0.0, interpval=0.0, kersum=0.0;

       for (int d=-wz; d<=wz; d++) {
	 sincz[d+wz] = kernelval((z-iz0+d),wz,sinckernelz);
       }
       for (int d=-wy; d<=wy; d++) {
	 sincy[d+wy] = kernelval((y-iy0+d),wy,sinckernely);
       }
       for (int d=-wx; d<=wx; d++) {
	 sincx[d+wx] = kernelval((x-ix0+d),wx,sinckernelx);
       }

       int xj, yj, zj;
       int numsum=0;
       for (int z1=iz0-wz; z1<=iz0+wz; z1++) {
	 zj=iz0-z1+wz;
	 for (int y1=iy0-wy; y1<=iy0+wy; y1++) {
	   yj=iy0-y1+wy;
	   for (int x1=ix0-wx; x1<=ix0+wx; x1++) {
	     if (v.in_bounds(x1,y1,z1)) {
	       xj=ix0-x1+wx;
	       float sincfac = sincx[xj] * sincy[yj] * sincz[zj];
	       convsum += v(x1,y1,z1) * sincfac;
	       kersum += sincfac;
	       numsum++;
	     } 
	   }
	 }
       }

       int numtot = (2*wz+1)*(2*wy+1)*(2*wx+1);
       int nummin = numtot/3;   // somewhat arbitrary choice
       if ( (numsum > nummin) && (fabs(kersum)>1e-9) ) {
	 interpval = convsum / kersum;
       } else {
	 return v.backgroundval();
       }
       return interpval;

     }
   


#ifndef NO_NAMESPACE
 }
#endif






